# -*- coding: utf-8 -*-
""" 
Log-likelihood functions used in model fitting.

"""
from __future__ import print_function, division

import numpy as np
from scipy.special import gamma, digamma


__all__ = ["StudTLogLike", "StudT2LogLike", "NormalLogLike", "Normal2LogLike",
           "JointLogLike"]

class LogLike:
    """
    Parameters
    ----------
    observed: numpy.ndarray
        Observed spectro-photometric SED of object.
    model: paintbox SED model
        SED model used in the modelling.
    obserr: numpy.ndarray, optional
        Uncertainties in the observed SED fitting to be used in the
        weighting of the log-likelihood.
    mask: numpy.ndarray, optional
        Boolean mask for the observed data. The mask uses the Python
        convention, such that False indicate points to be masked,
        and True indicate the points to be used.

    Attributes
    ----------
    parnames: list
        List with name of variables used in the evaluation of the
        log-likelihood.
    """

    def __init__(self, observed, model, obserr=None, mask=None):
        self.observed = observed
        self.model = model
        self.obserr = np.ones_like(self.observed) if obserr is None else obserr
        self.mask = np.full(len(self.observed), True) if mask is None \
                    else mask
        self._N = len(self.mask)
        self.parnames = self.model.parnames.copy()
        self._nparams = len(self.parnames)

class NormalLogLike(LogLike):
    r""" Normal loglikelihood for SED modeling.

    The normal log-likelihood is given by

    .. math::
       :nowrap:

       \begin{equation}
          \ln \mathcal{L}(y, \sigma|\theta)= -\frac{N}{2}\ln (2\pi)
          -\frac{1}{2}\sum_{i=1}^N \left (\frac{f(\theta)- y_i}{\sigma_i}
          \right )^2 - \frac{1}{2}\sum_{i=1}^{N}\ln \sigma_i^2
       \end{equation}

    where :math:`y` is the observed spectrum, :math:`\sigma` are the
    uncertainties, :math:`\theta` is the input vector of parameters and
    and :math:`f(\theta)` is the SED model.

    """
    __doc__ = __doc__ + LogLike.__doc__

    def __init__(self, observed, model, obserr=None, mask=None):
        super().__init__(observed, model, obserr=obserr, mask=mask)

    def __call__(self, theta):
        """ Calculation of the log-likelihood. """
        e_i = (self.model(theta) - self.observed)[self.mask]
        yerr = self.obserr[self.mask]
        LLF = - 0.5 * self._N * np.log(2 * np.pi) + \
              - 0.5 * np.sum(np.power(e_i / yerr, 2)) \
              - 0.5 * np.sum(np.log(yerr ** 2))
        return float(LLF)

    def gradient(self, theta):
        """ Gradient of the log-likelihood. """
        e_i = (self.model(theta) - self.observed)[self.mask]
        yerr = self.obserr[self.mask]
        g = self.model.gradient(theta)[:, self.mask]
        grad = - np.sum((e_i / np.power(yerr, 2.))[np.newaxis, :] * g,
                        axis=1)
        return grad

class Normal2LogLike(LogLike):
    r""" Variation of the normal log-likelihood with scaled errors.

    Uncertainties in the input spectrum may be under/ over estimated in some
    occassions, leading to under/over-estimated uncertainties in parameter
    estimation. This log-likelihood includes an extra parameter to
    scale the observed uncertainties by a multiplicative factor to increase
    the likelihood of the modeling. In this case, the log-likelihood is
    given by

    .. math::
        :nowrap:

        \begin{equation}
          \ln \mathcal{L}(y, \sigma|\theta, \eta)= -\frac{N}{2}\ln (2\pi)
          -\frac{1}{2}\sum_{i=1}^N \left (\frac{f(\theta)- y_i}{\eta \sigma_i}
          \right )^2 - \frac{1}{2}\sum_{i=1}^{N}\ln \eta^2\sigma_i^2
       \end{equation}

    where :math:`y` is the observed spectrum, :math:`\sigma` are the
    uncertainties, :math:`\theta` is the input vector of parameters and
    and :math:`f(\theta)` is the SED model. The multiplicative factor
    :math:`\eta` is appended to the parnames list.


    """
    __doc__ = __doc__ + LogLike.__doc__

    def __init__(self, observed, model, obserr=None, mask=None):
        super().__init__(observed, model, obserr=obserr, mask=mask)
        self.parnames += ["eta"]
        self._nparams += 1

    def __call__(self, theta):
        model = self.model(theta[:-1])
        if np.all(model) == 0:
            return -np.infty
        e_i = (model - self.observed)[self.mask]
        S = theta[-1]
        LLF = - 0.5 * self._N * np.log(2 * np.pi) + \
              - 0.5 * np.sum(np.power(e_i / (S * self.obserr[self.mask]), 2)) \
              - 0.5 * np.sum(np.log((S * self.obserr[self.mask]) ** 2))
        return float(LLF)

    def gradient(self, theta):
        e_i = (self.model(theta[:-1]) - self.observed)[self.mask]
        S = theta[-1]
        A = (e_i / np.power(S * self.obserr, 2.))[self.mask]
        B = self.model.gradient(theta[:-1])[self.mask]
        C = -np.sum(A[np.newaxis,:] * B, axis=1)[self.mask]
        grad = np.zeros(len(theta))
        grad[:-1] = C
        grad[-1] = - self._N / S + \
                   np.power(S, -3) * np.sum(np.power(e_i / self.obserr, 2))
        return grad

class StudTLogLike(LogLike):
    r"""Student's t-distribution log-likelihood.

    The Student's t-distribution log-likelihood allows for robust inference of
    parameters in models containing outliers. The log-likelihood is given by

    .. math::
        :nowrap:

        \begin{equation}
            \ln p(y, \sigma | \theta, \nu ) =
            N\log \left [ \frac{\Gamma\left (\frac{\nu + 1}{2}\right )}{\sqrt{
            \pi (\nu-2)}\Gamma\left (\frac{\nu}{2} \right )}\right ]
            -\frac{1}{2}\sum_{i=1}^{N}\log \sigma_{i}^2
            -\frac{\nu+1}{2}\sum_{i=1}^N \log \left [ 1 + \frac{\left (
            y_i - f(\theta)\right )^2}{\sigma_{i}^2 (\nu-2)} \right ]
        \end{equation}

    where :math:`y` is the observed spectrum, :math:`\sigma` are the
    uncertainties, :math:`\theta` is the input vector of parameters,
    :math:`f(\theta)` is the SED model, and :math:`\nu` is the
    degree-of-freedom parameter that controls the wings of the distribution,
    which is appended to the input parnames list.
    """
    __doc__ = __doc__ + LogLike.__doc__

    def __init__(self, observed, model, obserr=None, mask=None):
        super().__init__(observed, model, obserr=obserr, mask=mask)
        self.parnames += ["nu"]
        self._nparams += 1

    def __call__(self, theta):
        nu = theta[-1]
        e_i = self.model(theta[:-1])[self.mask] - self.observed[self.mask]
        x = 1. + np.power(e_i / self.obserr[self.mask], 2.) / (nu - 2)
        LLF = self._N * np.log(gamma(0.5 * (nu + 1)) /
                               np.sqrt(np.pi * (nu - 2)) / gamma(0.5 * nu)) \
              - 0.5 * (nu + 1) * np.sum(np.log(x)) \
              - 0.5 * np.sum(np.log(self.obserr ** 2)) # Constant
        return float(LLF)

    def gradient(self, theta):
        grad = np.zeros(self.model._nparams + 1)
        nu = theta[-1]
        # d loglike / d theta
        e_i = self.model(theta[:-1])[self.mask] - self.observed[self.mask]
        x = np.power(e_i / self.obserr[self.mask], 2.) / (nu - 2.)
        term1 = 1 / (1 + x)
        term2 = 2 * e_i / (self.obserr[self.mask] ** 2) / (nu - 2)
        term12 = term1 * term2
        sspgrad = self.model.gradient(theta[:-1])[:, self.mask]
        grad[:-1] = -0.5 * (nu + 1) * np.sum(term12[np.newaxis, :] *
                                             sspgrad, axis=1)
        # d loglike / d nu
        nuterm1 = 0.5 * self._N * digamma(0.5 * (nu + 1))
        nuterm2 = - 0.5 * self._N / (nu - 2)
        nuterm3 = -0.5 * self._N * digamma(0.5 * nu)
        nuterm4 = -0.5 * np.sum(np.log(1 + x))
        nuterm5 = 0.5 * (nu + 1) * np.power(nu - 2, -2) * \
                  np.sum(np.power(e_i / self.obserr, 2) * term1)
        grad[-1] = nuterm1 + nuterm2 + nuterm3 + nuterm4 + nuterm5
        return grad

class StudT2LogLike(LogLike):
    r"""Student's t-distribution log-likelihood with scaled uncertainties.

    Similar to the Normal2LogLike, this class extends the log-likelihood of
    the Student's t-distribution to include a term to scale the
    uncertainties to increase the likelihood as a way to compensate for
    under-over estimation of the observed uncertainties. In this case,
    the log-likelihood is given by

    .. math::
        :nowrap:

        \begin{equation}
            \ln p(y, \sigma | \theta, \eta, \nu ) =
            N\log \left [ \frac{\Gamma\left (\frac{\nu + 1}{2}\right )}{\sqrt{
            \pi (\nu-2)}\Gamma\left (\frac{\nu}{2} \right )}\right ]
            -\frac{1}{2}\sum_{i=1}^{N}\log \eta^2\sigma_{i}^2
            -\frac{\nu+1}{2}\sum_{i=1}^N \log \left [ 1 + \frac{\left (
            y_i - f(\theta)\right )^2}{\eta^2\sigma_{i}^2 (\nu-2)} \right ]
        \end{equation}

    where :math:`y` is the observed spectrum, :math:`\sigma` are the
    uncertainties, :math:`\theta` is the input vector of parameters,
    :math:`f(\theta)` is the SED model, :math:`\eta` is the parameter used
    to modify the scale of the uncertainties, and :math:`\nu` is the
    degree-of-freedom parameter that controls the wings of the distribution.
    Both :math:`\eta` and :math:`\nu` are appended to to input parameter list.
    """
    __doc__ = __doc__ + LogLike.__doc__

    def __init__(self, observed, model, obserr=None, mask=None):
        super().__init__(observed, model, obserr=obserr, mask=mask)
        self.parnames += ["eta", "nu"]
        self._nparams += 2

    def __call__(self, theta):
        S, nu = theta[-2:]
        e_i = self.model(theta[:-2])[self.mask] - self.observed[self.mask]
        x = 1. + np.power(e_i / S / self.obserr[self.mask], 2.) / (nu - 2)
        LLF = self._N * np.log(gamma(0.5 * (nu + 1)) /
                               np.sqrt(np.pi * (nu - 2)) / gamma(0.5 * nu)) \
              - 0.5 * (nu + 1) * np.sum(np.log(x)) \
              - 0.5 * np.sum(np.log((S * self.obserr[self.mask]) ** 2))
        return float(LLF)

    def gradient(self, theta):
        raise NotImplementedError("Gradients not supported for Studt2 "
                                  "loglikelyhood.")

class JointLogLike():
    def __init__(self, logp1, logp2):
        self.logp1 = logp1
        self.logp2 = logp2
        self.parnames = list(dict.fromkeys(logp1.parnames + logp2.parnames))
        self._idxs = []
        for parlist in [logp1.parnames, logp2.parnames]:
            idxs = []
            for p in parlist:
                idxs.append(self.parnames.index(p))
            self._idxs.append(np.array(idxs))

    def __call__(self, theta):
        return self.logp1(theta[self._idxs[0]]) + \
               self.logp2(theta[self._idxs[1]])


# -*- coding: utf-8 -*-
"""

Created on 30/10/2017

@Author: Carlos Eduardo Barbosa

Convolution routines for stellar populations analysis.

"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage.filters import gaussian_filter1d

def broad2lick(wl, intens, obsres, vel=0):
    """ Convolve spectra to match the Lick/IDS instrumental  resolution.

    Broad a given spectra to the Lick/IDS system resolution. The resolution
    of the IDS varies as function of the wavelength, and we use the mean
    interpolated values from the appendix of Worthey and Ottaviani 1997.

    ================
    Input parameters
    ================
    wl (array) :
        Wavelenght 1-D array in Angstroms.

    intens (array):
        Intensity 1-D array of Intensity, in arbitrary units. The
        lenght has to be the same as wl.

    obsres (float or array):
        Value of the observed resolution Full Width at Half Maximum (FWHM) in
        Angstroms.

    vel: float
        Recession velocity of the measured spectrum.

    =================
    Output parameters
    =================
    array_like
        The convolved intensity 1-D array.

    """
    c = 299792.458  # Speed of light in km/s
    dw = wl[1] - wl[0]
    if not isinstance(obsres, np.ndarray):
        obsres = np.ones_like(wl) * obsres
    wlick = np.array([2000., 4000., 4400., 4900., 5400., 6000., 8000.]) * \
            np.sqrt((1 + vel/c)/(1 - vel/c))
    lickres = np.array([11.5, 11.5, 9.2, 8.4, 8.4, 9.8, 9.8])
    flick = interp1d(wlick, lickres, kind="linear", bounds_error=False,
                         fill_value="extrapolate")
    fwhm_lick = flick(wl)
    fwhm_broad = np.sqrt(fwhm_lick**2 - obsres**2)
    sigma_b = fwhm_broad/ 2.3548 / dw
    intens2D = np.diag(intens)
    for i in range(len(sigma_b)):
        intens2D[i] = gaussian_filter1d(intens2D[i], sigma_b[i],
                      mode="constant", cval=0.0)
    return intens2D.sum(axis=0)

def broad2res(wave, intens, inres, outres):
    """ Convolve input spectra to a given resolution.

    Warning: This routine does not handle units, please keep consistency.

    Input parameters
    ----------------
    wave : np.array
        Linear wavelenght 1-D array in Angstroms.

    intens : np.array
        Intensity 1-D array of Intensity, in arbitrary units. The
        lenght has to be the same as wl.

    inres : float
        Input resolution (FWHM) of the spectrum

    outres: float
        Recession velocity of the measured spectrum.

    Output parameters
    -----------------
    array_like
        The convolved intensity 1-D array.

    """
    fwhm_diff = np.sqrt(outres**2 - inres**2)
    dw = wave[1] - wave[0]
    dsigma = fwhm_diff / 2.3548 / dw
    intens_broad = gaussian_filter1d(intens, dsigma,
                      mode="constant", cval=0.0)
    return intens_broad

if __name__ == "__main__":
    pass



""" Calculates the M/L for CvD models. """
import os
import glob

import numpy as np
import astropy.units as u
from astropy.table import Table, vstack
from astropy.io import fits
from tqdm import tqdm
from spectres import spectres
from scipy.ndimage.filters import gaussian_filter1d

from .disp2vel import disp2vel

def prepare_CvD18(ssp_files, wave, output, overwrite=False, sigma=100):
    """ Prepare templates for SSP models from Villaume et al. (2017).

    Parameters
    ----------
    ssp_files: list
        List containing the full path of SSP models to be processed.
    wave: np.array or astropy.Quantity
        Wavelength dispersion. Default units in Angstrom is assumed if
        wavelength is provided as as numpy array.
    output: str
        Name of the output file (a multi-extension FITS file)
    overwrite: bool (optional)
        Overwrite the output files if they already exist.
    sigma: float
        Velocity dispersion of the models, in km/s. Defaults to 100 km/s,
        the minimum resolution of the models.

    """
    if hasattr(wave, "unit"):
        wave = wave.to(u.Angstrom).value
    assert wave[0] >= 3501, "Minimum wavelength is 3501 Angstrom"
    assert wave[-1] <= 25000, "Maximum wavelength is 25000 Angstrom"
    assert sigma >= 100, "Minimum velocity dispersion is 100 km/s"
    if os.path.exists(output) and not overwrite:
        return
    nimf = 16
    imfs = 0.5 + np.arange(nimf) / 5
    x2s, x1s=  np.stack(np.meshgrid(imfs, imfs)).reshape(2, -1)
    velscale = int(sigma / 4)
    kernel_sigma = np.sqrt(sigma ** 2 - 100 ** 2) / velscale
    ssps, params = [], []
    for fname in tqdm(ssp_files, desc="Processing SSP files"):
        spec = os.path.split(fname)[1]
        T = float(spec.split("_")[3][1:])
        Z = float(spec.split("_")[4][1:-8].replace("p", "+").replace(
                    "m", "-"))
        for i, (x1, x2) in enumerate(zip(x1s, x2s)):
            params.append(Table([[Z], [T], [x1], [x2]],
                                names=["Z", "Age", "x1", "x2"]))
        data = np.loadtxt(fname)
        w = data[:,0]
        if sigma > 100:
            wvel = disp2vel(w, velscale)
        ssp = data[:, 1:].T
        if sigma <= 100:
            newssp = spectres(wave, w, ssp)
        else:
            ssp_rebin = spectres(wvel, w, ssp)
            ssp_broad = gaussian_filter1d(ssp_rebin, kernel_sigma,
                                          mode="constant", cval=0.0)
            newssp = spectres(wave, wvel, ssp_broad)

        ssps.append(newssp)
    ssps = np.vstack(ssps)
    params = vstack(params)
    hdu1 = fits.PrimaryHDU(ssps)
    hdu1.header["EXTNAME"] = "SSPS"
    params = Table(params)
    hdu2 = fits.BinTableHDU(params)
    hdu2.header["EXTNAME"] = "PARAMS"
    # Making wavelength array
    hdu3 = fits.BinTableHDU(Table([wave], names=["wave"]))
    hdu3.header["EXTNAME"] = "WAVE"
    hdulist = fits.HDUList([hdu1, hdu2, hdu3])
    hdulist.writeto(output, overwrite=True)
    return

def prepare_response_functions(rf_files, wave, outprefix, overwrite=False,
                               sigma=100):
    """ Prepare response functions from CvD models.

    Parameters
    ----------
    rf_files: str
        List of paths for the response function files.
    wave: np.array
        Wavelength dispersion for the output.
    outprefix: str
        First part of the name of the response function output files. The
        response functions are stored in different files for different
        elements, named "{}_{}.fits".format(outprefix, element).
    overwrite: bool (optional)
        Overwrite output.
    """
    if hasattr(wave, "unit"):
        wave = wave.to(u.Angstrom).value
    assert wave[0] >= 3501, "Minimum wavelength is 3501 Angstrom"
    assert wave[-1] <= 25000, "Maximum wavelength is 25000 Angstrom"
    assert sigma >= 100, "Minimum velocity dispersion is 100 km/s"
    # Read one spectrum to get name of columns
    with open(rf_files[0]) as f:
        header = f.readline().replace("#", "")
    fields = [_.strip() for _ in header.split(",")]
    fields[fields.index("C+")] = "C+0.15"
    fields[fields.index("C-")] = "C-0.15"
    fields[fields.index("T+")] = "T+50"
    fields[fields.index("T-")] = "T-50"
    fields = ["{}0.3".format(_) if _.endswith("+") else _ for _ in fields ]
    fields = ["{}0.3".format(_) if _.endswith("-") else _ for _ in fields]
    elements = set([_.split("+")[0].split("-")[0] for _ in fields if
                    any(c in _ for c in ["+", "-"])])
    outputs = ["{}_{}.fits".format(outprefix, element.replace("/", "_over_"))
               for element in elements]
    if all([os.path.exists(out) for out in outputs]) and not overwrite:
        return outputs
    signal = ["+", "-"]
    velscale = int(sigma / 4)
    kernel_sigma = np.sqrt(sigma ** 2 - 100 ** 2) / velscale
    rfsout = dict([(element, []) for element in elements])
    parsout = dict([(element, []) for element in elements])
    desc = "Preparing response functions"
    for i, fname in enumerate(tqdm(rf_files, desc=desc)):
        spec = os.path.split(fname)[1]
        T = float(spec.split("_")[2][1:])
        Z = float(spec.split("_")[3].split(".abun")[0][1:].replace(
            "p", "+").replace("m", "-"))
        data = np.loadtxt(fname)
        w = data[:, 0]
        data = data.T
        if sigma > 100:
            wvel = disp2vel(w, velscale)
            rebin = spectres(wvel, w, data)
            broad = gaussian_filter1d(rebin, kernel_sigma,
                                      mode="constant", cval=0.0)
            data = spectres(wave, wvel, broad).T

        else:
            data = spectres(wave, w, data).T
        fsun = data[:, 1]
        for element in elements:
            # Adding solar response
            p = Table([[Z], [T], [0.]], names=["Z", "Age", element])
            rfsout[element].append(np.ones(len(wave)))
            parsout[element].append(p)
            # Adding non-solar responses
            for sign in signal:
                name = "{}{}".format(element, sign)
                cols = [(i,f) for i, f in enumerate(fields) if f.startswith(
                    name)]
                for i, col in cols:
                    val = float("{}1".format(sign)) * float(col.split(sign)[1])
                    t = Table([[Z], [T], [val]], names=["Z", "Age", element])
                    parsout[element].append(t)
                    rf = data[:, i] / fsun
                    rfsout[element].append(rf)
    for element in elements:
        output = "{}_{}.fits".format(outprefix, element.replace("/", "_over_"))
        if os.path.exists(output) and not overwrite:
            continue
        rfdata = np.array(rfsout[element])
        params = vstack(parsout[element])
        hdu1 = fits.PrimaryHDU(rfdata)
        hdu1.header["EXTNAME"] = "SSPS"
        params = Table(params)
        hdu2 = fits.BinTableHDU(params)
        hdu2.header["EXTNAME"] = "PARAMS"
        # Making wavelength array
        hdu3 = fits.BinTableHDU(Table([wave], names=["wave"]))
        hdu3.header["EXTNAME"] = "WAVE"
        hdulist = fits.HDUList([hdu1, hdu2, hdu3])
        hdulist.writeto(output, overwrite=True)
    return outputs

def example():
    """ Example of application for the preparation of CvD models. """
    data_dir = "/home/kadu/Dropbox/SSPs/CvD18"
    ssps_dir = os.path.join(data_dir, "VCJ_v8")
    ssp_files = glob.glob(os.path.join(ssps_dir, "VCJ*.s100"))
    wave = np.linspace(4000, 20000, 2000)
    prepare_CvD18(ssp_files, wave, "CvD.fits", sigma=300, overwrite=True)
    rfs_dir = os.path.join(data_dir, "RFN_v3")
    rf_files = glob.glob(os.path.join(rfs_dir, "atlas_ssp*.s100"))
    outprefix = os.path.join(os.getcwd(), "rf")
    prepare_response_functions(rf_files, wave, "rf", sigma=300)


if __name__ == "__main__":
    example()


import numpy as np
import astropy.units as u


def disp2vel(wrange, velscale):
    """ Returns a log-rebinned wavelength dispersion with constant velocity.

    This code is an adaptation of pPXF's log_rebin routine, simplified to
    deal with the wavelength dispersion only.

    Parameters
    ----------
    wrange: list, np.array or astropy.Quantity
        Input wavelength dispersion range with two elements.

    velscale: float or astropy.Quantity
        Desired output velocity scale. Units are assumed to be km/s unless
        specified as an astropy.Quantity.

    """
    c = 299792.458  # Speed of light in km/s
    if isinstance(wrange, list):
        wrange = np.ndarray(wrange)
    wunits = wrange.unit if hasattr(wrange, "unit") else 1
    if hasattr(velscale, "unit"):
        velscale = velscale.to(u.km/u.s).value
    veldiff = np.log(np.max(wrange) / np.min(wrange)) * c
    n = veldiff / velscale
    m = int(n)
    dv = 0.5 * (n-m) * velscale
    v = np.arange(0, m * velscale, velscale) + dv
    w = wrange[0] * np.exp(v / c)
    return w * wunits


# -*- coding: utf-8 -*-
"""

Created on 16/05/16

@author: Carlos Eduardo Barbosa

Program to calculate lick indices

"""

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline
import astropy.units as u
from astropy import constants

class Lick():
    """ Class to measure Lick indices.

    Computation of the Lick indices in a given spectrum. Position of the
    passbands are determined by redshifting the position of the bands
    to the systemic velocity of the galaxy spectrum.

    =================
    Input parameters:
    =================
        wave (array):
            Wavelength of the spectrum given.

        galaxy (array):
            Galaxy spectrum in arbitrary units.

        bands0 (array) :
            Definition of passbands for Lick indices at rest
            wavelengths. Units should be consistent with wavelength array.

        vel (float, optional):
            Systemic velocity of the spectrum in km/s. Defaults to zero.

        dw (float, optinal):
            Extra wavelength to be considered besides
            bands for interpolation. Defaults to 2 wavelength units.

    ===========
    Attributes:
    ===========
        bands (array):
            Wavelengths of the bands after shifting to the
            systemic velocity of the galaxy.

    """
    def __init__(self, wave, galaxy, bands0, vel=None, dw=None, units=None):
        self.galaxy = galaxy
        self.wave = wave.to(u.AA).value
        self.vel = vel
        self.bands0 = bands0.to(u.AA).value
        if dw is None:
            self.dw = 2
        if vel is None:
            self.vel = 0 * u.km / u.s
        self.units = units if units is not None else \
                                                np.ones(len(self.bands0)) * u.AA
        ckms = constants.c.to("km/s")
        # c = 299792.458 # Speed of light in km/s
        self.bands = self.bands0 * np.sqrt((1 + self.vel.to("km/s")/ckms)
                    /(1 - self.vel.to("km/s")/ckms))

    def classic_integration(self):
        """ Calculation of Lick indices using spline integration.

        ===========
        Attributes:
        ===========
            R (array):
                Raw integration values for the Lick indices.

            Ia (array):
                Indices measured in equivalent widths.

            Im (array):
                Indices measured in magnitudes.

            classic (array):
                Indices measured according to the conventional
                units mixturing equivalent widths and magnitudes.
        """
        self.R = np.zeros(self.bands._shape[0])
        self.Ia = np.zeros_like(self.R)
        self.Im = np.zeros_like(self.R)
        for i, w in enumerate(self.bands):
            condition = np.array([w[0]-self.dw > self.wave[0],
                                 w[-1]+self.dw < self.wave[-1]])
            if not np.all(condition):
                self.R[i] = np.nan
                self.Ia[i] = np.nan
                self.Im[i] = np.nan
                continue
            # Defining indices for each section
            idxb = np.where(((self.wave > w[0] - self.dw) &
                                 (self.wave < w[1] + self.dw)))
            idxr = np.where(((self.wave > w[4] - self.dw) &
                                (self.wave < w[5] + self.dw)))
            idxcen = np.where(((self.wave > w[2] - self.dw) &
                                (self.wave < w[3] + self.dw)))
            # Defining wavelenght samples
            wb = self.wave[idxb]
            wr = self.wave[idxr]
            wcen = self.wave[idxcen]
            # Defining intensity samples
            fb = self.galaxy[idxb]
            fr = self.galaxy[idxr]
            fcen = self.galaxy[idxcen]
            # Interpolation functions for pseudocontinuum
            sb = InterpolatedUnivariateSpline(wb, fb)
            sr = InterpolatedUnivariateSpline(wr, fr)
            # Calculating the mean fluxes for the pseudocontinuum
            fp1 = sb.integral(w[0], w[1]) / (w[1] - w[0])
            fp2 = sr.integral(w[4], w[5]) / (w[5] - w[4])
            # Making pseudocontinuum vector
            x1 = (w[0] + w[1])/2.
            x2 = (w[4] + w[5])/2.
            fc = fp1 + (fp2 - fp1)/ (x2 - x1) * (wcen - x1)
            # Calculating indices
            ffc = InterpolatedUnivariateSpline(wcen, fcen/fc/(w[3]-w[2]))
            self.R[i] =  ffc.integral(w[2], w[3])
            self.Ia[i] = (1 - self.R[i]) * (w[3]-w[2])
            self.Im[i] = -2.5 * np.log10(self.R[i])
        self.Ia = self.Ia * u.AA
        self.Im = self.Im * u.mag
        idx = np.where([_ == u.Unit("mag") for _ in self.units])[0]
        self.classic = np.copy(self.Ia)
        self.classic[idx] = self.Im[idx]
        return

def bands_shift(bands, vel):
    c = 299792.458  # Speed of light in km/s
    return  bands * np.sqrt((1 + vel/c)/(1 - vel/c))

if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

Created on 04/05/16

@author: Carlos Eduardo Barbosa

Calculates and plot the spectral resolution of MUSE.

"""
from __future__ import print_function

import numpy as np
from scipy.ndimage.filters import gaussian_filter1d

def broad2res(w, flux, obsres, outres, fluxerr=None):
    """ Broad resolution of observed spectra to a given resolution.

    Input Parameters
    ----------------
    w : np.array
        Wavelength array

    flux: np.array
        Spectrum to be broadened to the desired resolution.

    obsres : float or np.array
        Observed wavelength spectral resolution FWHM.

    outres: float
        Resolution FWHM  of the spectra after the broadening.

    fluxerr: np.array
        Spectrum errors whose uncertainties are propagated

    Output parameters
    -----------------
    np.array:
        Broadened spectra.

    """
    dws = np.diff(w)
    dw = np.median(dws)
    assert np.all(np.isclose(dws, dw)), \
        "Wavelength dispersion has to be constant!"
    sigma_diff = np.sqrt(outres ** 2 - obsres ** 2) / 2.3548 / dw
    diag = np.diag(flux)
    for j in range(len(w)):
        diag[j] = gaussian_filter1d(diag[j], sigma_diff[j], mode="constant",
                                 cval=0.0)
    newflux = diag.sum(axis=0)
    if fluxerr is None:
        return newflux
    errdiag = np.diag(fluxerr)
    for j in range(len(w)):
        errdiag[j] = gaussian_filter1d(errdiag[j]**2, sigma_diff[j],
                                       mode="constant", cval=0.0)
    newfluxerr = np.sqrt(errdiag.sum(axis=0))
    return newflux, newfluxerr


# -*- coding: utf-8 -*-
"""

Basic SED classes for handling stellar population and emission lines
based on precomputed templates and polynomials.

"""
from __future__ import print_function, division

import numpy as np
from scipy.interpolate import RegularGridInterpolator
from scipy.special import legendre

from .operators import CompositeSED

__all__ = ["ParametricModel", "NonParametricModel", "Polynomial"]

class ParametricModel():
    """
    Interpolation of SED model templates parametrically.

    This class allows the linear interpolation of SED templates, such as SSP
    models, based on a table of parameters and their SEDs.

    Warning: The linear interpolation is currently based on
    scipy.RegularGridInterpolator for better performance with large number of
    input models, thus the input data must be in the form of a regular grid.

    Attributes
    ----------
    parnames: list
        Name of the variables of the SED model.

    """
    def __init__(self, wave, params, data):
        """
        Parameters
        ----------
        wave: ndarray, Quantity
            Wavelenght array of the model.
        params: astropy.table.Table
            Table with parameters of the models.
        data: 2D ndarray
            The SED templates with dimensions (len(params), len(wave))

        """
        self.wave = wave
        self.params = params
        self.data = data
        self.parnames = self.params.colnames.copy()
        self._n = len(wave)
        self._nparams = len(self.parnames)
        # Interpolating models
        x = self.params.as_array()
        pdata = x.view((x.dtype[0], len(x.dtype.names)))
        nodes = []
        for param in self.parnames:
            x = np.unique(self.params[param]).data
            nodes.append(x)
        coords = np.meshgrid(*nodes, indexing='ij')
        dim = coords[0].shape + (self._n,)
        data = np.zeros(dim)
        with np.nditer(coords[0], flags=['multi_index']) as it:
            while not it.finished:
                multi_idx = it.multi_index
                x = np.array([coords[i][multi_idx] for i in range(len(coords))])
                idx = (pdata == x).all(axis=1).nonzero()[0]
                data[multi_idx] = self.data[idx]
                it.iternext()
        self._interpolator = RegularGridInterpolator(nodes, data,
                             bounds_error=False, fill_value=0)
        ########################################################################
        # Get grid points to handle derivatives
        inner_grid = []
        thetamin = []
        thetamax = []
        for par in self.parnames:
            thetamin.append(np.min(self.params[par].data))
            thetamax.append(np.max(self.params[par].data))
            inner_grid.append(np.unique(self.params[par].data)[1:-1])
        self._thetamin = np.array(thetamin)
        self._thetamax = np.array(thetamax)
        self._inner_grid = inner_grid

    def __call__(self, theta):
        """ Call for interpolated model at a given point theta.

        Parameters
        ----------
        theta: ndarray
            Point where the model is computed, with parameters in
            the same order of parnames. Points outside of the convex hull of
            the models are set to zero.

        Returns
        -------
        SED model at location theta.
        """
        return self._interpolator(theta)[0]

    def __add__(self, o):
        """ Addition between two SED components. """
        return CompositeSED(self, o, "+")

    def __mul__(self, o):
        """  Multiplication between two SED components. """
        return CompositeSED(self, o, "*")

    def gradient(self, theta, eps=1e-6):
        """ Gradient of models at a given point theta.

        Gradients are computed with simple finite difference. If the input
        point is among the points used for interpolation, the gradient is not
        defined, returning zero instead.

        Parameters
        ----------
        theta: ndarray
            Point where the gradient of the model is computed,
            with parameters in the same order of parnames. Points outside
            of the convex hull of the models are set to zero.
        eps: float or ndarray, optional
            Step used in the finite difference calculation. Default is 1e-6.

        """
        # Clipping theta to avoid border problems
        theta = np.maximum(theta, self._thetamin + 2 * eps)
        theta = np.minimum(theta, self._thetamax - 2 * eps)
        grads = np.zeros((self._nparams, self._n))
        for i,t in enumerate(theta):
            epsilon = np.zeros(self._nparams)
            epsilon[i] = eps
            # Check if data point is in inner grid
            in_grid = t in self._inner_grid[i]
            if in_grid:
                continue
            else:
                tp = theta + epsilon
                tm = theta - epsilon
                grads[i] = (self.__call__(tp) - self.__call__(tm)) / (2 * eps)
        return grads

class NonParametricModel():
    """
    Weighted linear combination of SED models.

    This class allows the combination of a set of templates based on
    different weights.

    Attributes
    ----------
    parnames: list
        Name of the templates.

    """
    def __init__(self, wave, templates, names=None):
        """ 
        Parameters
        ----------
        wave: ndarray, Quantity
            Common wavelenght array of all templates.
        templates: 2D ndarray
            SED models with dimensions (N, len(wave)), where N=number of
            templates.
        names: list
            Name of the templates. Defaults to [temp, ..., tempN]
        """
        self.wave = wave
        self.templates = np.atleast_2d(templates)
        self._nparams = len(self.templates)
        self._n = len(templates)
        names = ["temp{}".format(n+1) for n in range(self._n)] if \
                 names is None else names
        self.parnames = names
        self._nparams = len(self.parnames)

    def __call__(self, theta):
        """ Returns the dot product of a vector theta with the templates.

        Parameters
        ----------
        theta: ndarray
            Vector with weights of the templates.

        Returns
            Dot product of theta with templates.
        """
        return np.dot(theta, self.templates)

    def __add__(self, o):
        """ Addition between two SED components. """
        return CompositeSED(self, o, "+")

    def __mul__(self, o):
        """  Multiplication between two SED components. """
        return CompositeSED(self, o, "*")

    def gradient(self, theta):
        """ Gradient of the dot product with weights theta.

        This routine returns simply the templates, but it has an argument
        theta only to keep calls consistently across different SED
        components.
        """
        return self.templates

class Polynomial():
    """
    Polynomial SED component.

    This class produces Legendre polynomials that can be either added or
    multiplied with other SED components.

    Attributes
    ----------
    wave: ndarray, Quantity
        Wavelength array of the polynomials
    degree: int
        Order of the Legendre polynomial.
    poly: 2D ndarray
        Static Legendre polynomials array used in the calculations.
    parnames: list
        List with the name of the individual polynomials, set to
        [p0, p1, ..., pdegree] at initialization.

    """

    def __init__(self, wave, degree):
        """
        Parameters
        ----------
        wave: ndarray, Quantity
            Wavelength array of the polynomials
        degree: int
            Order of the Legendre polynomial.fg
        """
        self.wave = wave
        self.degree = degree
        self._x = 2 * ((self.wave - self.wave.min()) /
                       (self.wave.max() - self.wave.min()) - 0.5)
        self.poly = np.zeros((self.degree + 1, len(self._x)))
        for i in np.arange(self.degree + 1):
            self.poly[i] = legendre(i)(self._x)
        self.parnames = ["p{}".format(i) for i in np.arange(degree+1)]
        self._nparams = len(self.parnames)

    def __call__(self, theta):
        """ Calculation of the polynomial with weights theta. """
        return np.dot(theta, self.poly)

    def __add__(self, o):
        """ Addition between two SED components. """
        return CompositeSED(self, o, "+")

    def __mul__(self, o):
        """  Multiplication between two SED components. """
        return CompositeSED(self, o, "*")

    def gradient(self, theta):
        """ Gradient of the polynomial with weights theta. """
        return self.poly


# -*- coding: utf-8 -*-
"""
Miscelaneous classes for combination and modification of other paintbox model
classes.
"""
from __future__ import print_function, division

import numpy as np
import astropy.constants as const
from scipy.ndimage import convolve1d
from spectres import spectres

__all__ = ["LOSVDConv", "Resample", "CompositeSED", "Constrain"]

class LOSVDConv():
    """ Convolution of a SED model with the LOS Velocity Distribution.

    This class allows the convolution of a SED model with the line--of-sight
    velocity distribution. This class requires that the input SED model has a
    logarithmic-scaled wavelength dispersion. Currently, this operation only
    supports LOSVDs with two moments (velocity and velocity dispersion).
    
    Attributes
    ----------
    velscale: Quantity
        Velocity scale of the wavelength array.
    parnames: list
        Updated list of parameters, including the input SED object parameter
        names and the LOSVD parameters.
    wave: numpy.ndarray, astropy.quantities.Quantity
        Wavelength array.
    """
    def __init__(self, obj, losvdpars=None, vunit="km/s"):
        """
        Parameters
        ----------
        obj: SED model
            Input paintbox SED model to be convolved with the LOSVD.
        losvdpars: list
            Name of the LOSVD parameters appended to the input paranames.
            Defaults to [V, sigma].
        vunit: str
            Units used for velocity variables. Defaults to km/s.

        """
        self.obj = obj
        self.wave = obj.wave
        # Check if velocity scale is constant
        velscale = np.diff(np.log(self.wave) * const.c.to(vunit))
        msg = "LOSVD convolution requires wavelength array with constant " \
              "velocity scale."
        assert np.all(np.isclose(velscale, velscale[0])), msg
        self.velscale = velscale[0]
        self._v = self.velscale.value
        self.losvdpars = ["V", "sigma"] if losvdpars is None else losvdpars
        self.parnames = obj.parnames + self.losvdpars
        self._nparams = len(self.parnames)
        self._shape = (self._nparams, len(self.wave))

    def _kernel_arrays(self, p):
        """ Produces kernels used in the convolution. """
        x0, sigx = p / self._v
        dx = int(np.ceil(np.max(abs(x0) + 5 * sigx)))
        n = 2 * dx + 1
        x = np.linspace(-dx, dx, n)
        y = (x - x0) / sigx
        y2 = np.power(y, 2.)
        k = np.exp(-0.5 * y2) / (sigx * np.sqrt(2 * np.pi))
        return y, k

    def __call__(self, theta):
        """ Performs convolution of input model with LOSVD."""
        z = self.obj(theta[:-2])
        y, k = self._kernel_arrays(theta[-2:])
        return convolve1d(z, k)

    def gradient(self, theta):
        """ Gradient of the convolved model. """
        p1 = theta[:-2]
        p2 = theta[-2:]
        grad = np.zeros(self._shape)
        model = self.obj(theta[:-2])
        modelgrad = self.obj.gradient(p1)
        y, k = self._kernel_arrays(p2)
        for i in range(len(modelgrad)):
            grad[i] = convolve1d(modelgrad[i], k)
        grad[-2] = convolve1d(model, y * k / p2[1])
        grad[-1] = convolve1d(model, (y * y - 1.) * k / p2[1])
        return grad

    def __add__(self, o):
        """ Addition of this output model with other SED components. """
        return CompositeSED(self, o, "+")

    def __mul__(self, o):
        """ Multiplication of this output model with other SED components. """
        return CompositeSED(self, o, "*")

class Resample():
    """ Resampling of SED model to a new wavelength dispersion.

    The resample can be performed to arbitrary dispersions based on the
    'spectres <https://spectres.readthedocs.io/en/latest/>'_ package.

    Attributes
    ----------
    parnames: list
        List of parameter names.
    wave: numpy.ndarray, astropy.quantities.Quantity
        Wavelength array.
    """
    def __init__(self, wave, obj):
        """
        Parameters
        ----------
        wave: ndarray, Quantity
            New wavelength array of the SED model.
        obj: SED model
            SED model to be resampled.
        """
        self.obj = obj
        self._inwave= self.obj.wave
        self.wave = wave
        self.parnames = self.obj.parnames
        self._nparams = len(self.parnames)

    def __call__(self, theta):
        """ Performs the resampling. """
        model = self.obj(theta)
        rebin = spectres(self.wave, self._inwave, model)
        return rebin

    def gradient(self, theta):
        """ Calculation the the gradient of the resampled model. """
        grads = self.obj.gradient(theta)
        grads = spectres(self.wave, self._inwave, grads)
        return grads

    def __add__(self, o):
        """ Addition of this model with other SED models. """
        return CompositeSED(self, o, "+")

    def __mul__(self, o):
        """ Multiplication of this model with other SED models. """
        return CompositeSED(self, o, "*")
    
class CompositeSED():
    """
    Combination of SED models.

    The CompositeSED class allows the combination of any number of SED model
    components using addition and / or multiplication, as long as the input
    classes have the same wavelength dispersion.

    Attributes
    ----------
    parnames: list
        The new parnames list is a concatenation of the input SED models.
    wave: numpy.ndarray, astropy.quantities.Quantity
        Wavelength array.
    """
    def __init__(self, o1, o2, op):
        """
        Parameters
        ----------
        o1, o2: SED model components
            Input SED models to be combined either by multiplication or
            addition.
        op: str
            Operation of the combination, either "+" or "*".
        """
        msg = "Components with different wavelenghts cannot be combined!"
        assert np.all(o1.wave == o2.wave), msg
        self.__op = op
        msg = "Operations allowed in combination of SED components are + and *."
        assert self.__op in ["+", "*"], msg
        self.o1 = o1
        self.o2 = o2
        self.wave = self.o1.wave
        self.parnames = self.o1.parnames + self.o2.parnames
        self._nparams = len(self.parnames)
        self._grad_shape = (self._nparams, len(self.wave))

    def __call__(self, theta):
        """ SED model for combined components at point theta. """
        theta1 = theta[:self.o1._nparams]
        theta2 = theta[self.o1._nparams:]
        if self.__op == "+":
            return self.o1(theta1) + self.o2(theta2)
        elif self.__op == "*":
            return self.o1(theta1) * self.o2(theta2)
    
    def gradient(self, theta):
        """ Gradient of the combined SED model at point theta. """
        n = self.o1._nparams
        theta1 = theta[:n]
        theta2 = theta[n:]
        grad = np.zeros(self._grad_shape)
        if self.__op == "+":
            grad[:n, :] = self.o1.gradient(theta1)
            grad[n:, :] = self.o2.gradient(theta2)
        elif self.__op == "*":
            grad[:n, :] = self.o1.gradient(theta1) * self.o2(theta2)
            grad[n:, :] = self.o2.gradient(theta2) * self.o1(theta1)
        return grad

    def __add__(self, o):
        """ Addition of SED components. """
        return CompositeSED(self, o, "+")

    def __mul__(self, o):
        """ Multiplication of SED components. """
        return CompositeSED(self, o, "*")

class Constrain():
    """ Constrain parameters of an SED model.

    The combination of SED models may result in models with repeated
    parameters at different locations of the parnames list. This class allows
    the simplification of the input model by finding and constraining all
    instances of repeated parameters to the same value.

    Attributes
    ----------
    parnames: list
        The new parnames list is a concatenation of the input SED models,
        simplified in relation to the input model.
    wave: numpy.ndarray, astropy.quantities.Quantity
        Wavelength array.

    Methods
    -------
    __call__(theta)
        Returns the SED model according the parameters given in theta.

    """
    def __init__(self, sed):
        """


        """
        self.sed = sed
        self.parnames = list(dict.fromkeys(sed.parnames))
        self.wave = self.sed.wave
        self._nparams = len(self.parnames)
        self._shape = len(self.sed.parnames)
        self._idxs = {}
        for param in self.parnames:
            self._idxs[param] = np.where( \
                                np.array(self.sed.parnames) == param)[0]

    def __call__(self, theta):
        """ Calculates the constrained model. """
        t = np.zeros(self._shape)
        for param, val in zip(self.parnames, theta):
            t[self._idxs[param]] = val
        return self.sed(t)

    def gradient(self, theta):
        raise NotImplementedError


# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *   # noqa
# ----------------------------------------------------------------------------

__all__ = []
# from .example_mod import *   # noqa
# Then you can be explicit to control what ends up in the namespace,

# or you can keep everything from the subpackage with the following instead
# __all__ += example_mod.__all__

from .operators import *
from .sed import *
from .extlaws import *
from .likelihoods import *
    # from .interfaces import *


# -*- coding: utf-8 -*-
""" 
Extinction laws used for dust attenuation.

"""
from __future__ import print_function, division

import numpy as np
import astropy.units as u

from .operators import CompositeSED

__all__ = ["CCM89", "C2000"]

class CCM89():
    r""" Cardelli, Clayton and Mathis (1989) extinction law.

    The extinction laws are calculated using a dust screen model, returning
    the ratio

    .. math::
        :nowrap:

        \begin{equation}
            \frac{f_\lambda}{f_\lambda^0}= 10^{-0.4 A_V
            \left (a(x) + b(x) / R_V\right )}
        \end{equation}

    where :math:`x=1/\lambda`, :math:`a(x)` and :math:`b(x)` are fixed
    polynomials, :math:`A_V` is the total extinction, and :math:`R_V` is the
    total-to-selective extinction ratio.
    """
    def __init__(self, wave, unit=None):
        """
        Parameters
        ----------
        wave: numpy.ndarray, astropy.quantities.Quantity
            Wavelength array
        unit: str, optional
            Units can be specified in the form of a string. Default is
            Angstrom. This parameter is only used if the input wavelength
            array is not an astropy.quantities.Quantity.

        Attributes
        ----------
        parnames: list
            Name of the free parameters (Av, Rv) of the extinction law.


        """
        if hasattr(wave, "unit"):
            self.wave = wave.value
            self.unit = wave.unit
        else:
            self.wave = wave
            self.unit = u.AA if unit is None else unit
        x = 1 / (self.wave * self.unit).to(u.micrometer).value
        self.parnames = ["Av", "Rv"]
        self._nparams = 2

        def _anir(x):
            return 0.574 * np.power(x, 1.61)

        def _bnir(x):
            return -0.527 * np.power(x, 1.61)

        def _aopt(x):
            y = x - 1.82
            return 1 + 0.17699 * y - 0.50447 * np.power(y, 2) \
                   - 0.02427 * np.power(y, 3) + 0.7208 * np.power(y, 4) \
                   + 0.0197 * np.power(y, 5) - 0.7753 * np.power(y, 6) \
                   + 0.32999 * np.power(y, 7)

        def _bopt(x):
            y = x - 1.82
            return 1.41338 * y + 2.28305 * np.power(y, 2) + \
                   1.07233 * np.power(y, 3) - 5.38434 * np.power(y, 4) - \
                   0.62251 * np.power(y, 5) + 5.30260 * np.power(y, 6) - \
                   2.09002 * np.power(y, 7)

        def _auv(x):
            Fa = - 0.04473 * np.power(x - 5.9, 2) - 0.009779 * np.power(x - 5.9,
                                                                        3)
            a = 1.752 - 0.316 * x - 0.104 / (np.power(x - 4.67, 2) + 0.341)
            return np.where(x < 5.9, a, a + Fa)

        def _buv(x):
            Fb = 0.2130 * np.power(x - 5.9, 2) + 0.1207 * np.power(x - 5.9, 3)
            b = -3.090 + 1.825 * x + 1.206 / (np.power(x - 4.62, 2) + 0.263)
            return np.where(x < 5.9, b, b + Fb)

        nir = (0.3 <= x) & (x <= 1.1)
        optical = (1.1 < x) & (x <= 3.3)
        uv = (3.3 < x) & (x <= 8)
        self._a = np.where(nir, _anir(x), np.where(optical, _aopt(x),
                                                   np.where(uv, _auv(x), 0)))
        self._b = np.where(nir, _bnir(x), np.where(optical, _bopt(x),
                                                   np.where(uv, _buv(x), 0)))

    def __call__(self, theta):
        """ Returns the dust screen model attenuation.

        Parameters
        ----------
        theta: numpy.ndarray
            Array with values of Av and Rv.
        """
        return np.power(10, -0.4 * theta[0] * (self._a + self._b / theta[1]))

    def __add__(self, o):
        """ Addition between two SED components. """
        return CompositeSED(self, o, "+")

    def __mul__(self, o):
        """  Multiplication between two SED components. """
        return CompositeSED(self, o, "*")

    def gradient(self, theta):
        """ Gradient of the extinction law.

        Parameters
        ----------
        theta: numpy.ndarray
            Array with values of Av and Rv.
        """
        grad = np.zeros((2, len(self.wave)))
        A = self.__call__(theta)
        grad[0] = -0.4 * np.log(10) * (self._a + self._b / theta[1]) * A
        grad[1] = 0.4 * np.log(10) * theta[0] * self._b * \
                  np.power(theta[1], -2) * A
        return grad

class C2000():
    r""" Calzetti et al. (2000) extinction law.

    The extinction laws are calculated using a dust screen model, returning
    the ratio

    .. math::
        :nowrap:

        \begin{equation}
        \frac{f_\lambda}{f_\lambda^0}= 10^{-0.4 A_V \left (1 +
                                       \kappa_\lambda / R_V\right )}
        \end{equation}
    """
    def __init__(self, wave, unit=None):
        """
        Parameters
        ----------
        wave: numpy.ndarray, astropy.quantities.Quantity
            Wavelength array
        unit: str, optional
            Units can be specified in the form of a string. Defalt is Angstrom.

        """
        if hasattr(wave, "unit"):
            self.wave = wave.value
            self.unit = wave.unit
        else:
            self.wave = wave
            self.unit = u.AA if unit is None else u.Quantity(unit)
        x = 1 / (self.wave * self.unit).to(u.micrometer).value
        self._kappa = np.where(self.wave > 0.63 * u.micrometer,
                               2.659 * (-1.857 + 1.040 * x), \
                               2.659 * (-2.156 + 1.509 * x - 0.198 * x * x
                                       + 0.011 * (x * x * x)))

    def __call__(self, theta):
        """ Returns the dust screen model attenuation.

        Parameters
        ----------
        theta: numpy.ndarray
            Array with values of Av and Rv.
        """
        return np.power(10, -0.4 * theta[0] * (1. + self._kappa / theta[1]))

    def __add__(self, o):
        """ Addition between two SED components. """
        return CompositeSED(self, o, "+")

    def __mul__(self, o):
        """  Multiplication between two SED components. """
        return CompositeSED(self, o, "*")

    def gradient(self, theta):
        """ Gradient of the extinction law.

        Parameters
        ----------
        theta: numpy.ndarray
            Array with values of Av and Rv.
        """
        grad = np.zeros((2, len(self.wave)))
        A = self.__call__(theta)
        # dAw / dAv
        grad[0] = A * np.log(10) * (-0.4 * (1. + self._kappa / theta[1]))
        # dAw / dRv
        grad[1] = A * 0.4 * theta[0] * self._kappa * np.log(10) * \
                  np.power(theta[1], -2.)
        return grad


# -*- coding: utf-8 -*-
"""

Created on 27/11/19

Author : Carlos Eduardo Barbosa

"""
from __future__ import print_function, division

import numpy as np
import theano.tensor as tt

__all__ = ["TheanoLogLikeInterface"]

class TheanoLogLikeInterface(tt.Op):
    """
    Produces a theano.tensor operator for the data log-likelihood to be used
    with pymc3.

    Parameters
    ----------
    likelyhood: paintbox.Likelihood object

    """
    itypes = [tt.dvector] # expects a vector of parameter values when called
    otypes = [tt.dscalar] # outputs a single scalar value (the log likelihood)

    def __init__(self, likelyhood):
        self.likelihood = likelyhood
        # initialise the gradient Op (below)
        self.logpgrad = _LogLikeGrad(self.likelihood)

    def perform(self, node, inputs, outputs):
        theta, = inputs
        logl = self.likelihood(theta)
        outputs[0][0] = np.array(logl) # output the log-likelihood

    def grad(self, inputs, g):
        theta, = inputs  # our parameters
        return [g[0] * self.logpgrad(theta)]



class _LogLikeGrad(tt.Op):

    """
    This Op will be called with a vector of values and also return a vector of
    values - the gradients in each dimension.
    """
    itypes = [tt.dvector]
    otypes = [tt.dvector]

    def __init__(self, likelihood):
        """
        Initialise with various things that the function requires. Below
        are the things that are needed in this particular example.

        Parameters
        ----------
        loglike:
            The log-likelihood (or whatever) function we've defined
        """

        # add inputs as class attributes
        self.likelihood = likelihood

    def perform(self, node, inputs, outputs):
        theta, = inputs
        # calculate gradients
        grads = self.likelihood.gradient(theta)
        outputs[0][0] = grads